@@ -22,7 +22,7 @@
 #include <libnftnl/rule.h>
 
 #ifndef NFT_META_MAX
-#define NFT_META_MAX (NFT_META_SDIFNAME + 1)
+#define NFT_META_MAX (NFT_META_BRI_BROUTE + 1)
 #endif
 
 struct nftnl_expr_meta {
@@ -168,6 +168,7 @@ static const char *meta_key2str_array[NFT_META_MAX] = {
 	[NFT_META_TIME_HOUR]	= "hour",
 	[NFT_META_SDIF]		= "sdif",
 	[NFT_META_SDIFNAME]	= "sdifname",
+	[NFT_META_BRI_BROUTE]	= "broute",
 };
 
 static const char *meta_key2str(uint8_t key)
